//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"errors"
	"io/ioutil"
	"log"
	"runtime"
	"sync"

	"gitlab.servus.at/autoradio/tank/store"
)

var (
	ErrNotImplemented = errors.New("not implemented")
	ErrNotFound       = errors.New("not found")
	ErrTimeout        = errors.New("timeout")
)

type Importer struct {
	conf     *Config
	infoLog  *log.Logger
	errLog   *log.Logger
	dbgLog   *log.Logger
	wgWorker sync.WaitGroup
	work     chan *Job
	jobs     *jobInventory
}

func (im *Importer) CreateJob(group string, id uint64, src, user, refID string) (*Job, error) {
	// TODO: implement this
	// sanity check inputs:
	//   - parse src: valid URI?
	//   - does fileID exist and does it belong to group?
	//   - file.Source.Import.State == new?
	//
	// add job to inventory -> return existing job if it already exists
	// job.Start() will enqueue the job
	return nil, ErrNotImplemented
}

func (im *Importer) GetJob(group string, id uint64) (*Job, error) {
	// TODO: pass this on to job inventory
	return nil, ErrNotImplemented
}

func (im *Importer) ListJobs(group string) ([]*Job, error) {
	// TODO: pass this on to job inventory
	return nil, ErrNotImplemented
}

// handle subscriptions to new jobs-of-group -> pass on to job inventory

func (im *Importer) deleteJob(group string, id uint64) ([]*Job, error) {
	// TODO: pass this on to job inventory
	return nil, ErrNotImplemented
}

func (im *Importer) runWorker(idx int) {
	defer im.dbgLog.Printf("importer: worker(%d) has stopped", idx)

	im.dbgLog.Printf("importer: worker(%d) is running", idx)
	for job := range im.work {
		select {
		case <-job.ctx.Done():
			continue
		default:
		}
		im.infoLog.Printf("importer: worker(%d) starting job(%s/%d) ...", idx, job.Group, job.ID)

		err := ErrTimeout
		select {
		case <-job.ctx.Done():
		case err = <-job.start():
		}
		if err != nil {
			im.errLog.Printf("importer: worker(%d) job(%s/%d) failed: %s", idx, job.Group, job.ID, err.Error())
		} else {
			im.infoLog.Printf("importer: worker(%d) successfully completed job(%s/%d)", idx, job.Group, job.ID)
		}
	}
}

func NewImporter(conf Config, st *store.Store, infoLog, errLog, dbgLog *log.Logger) (im *Importer, err error) {
	if infoLog == nil {
		infoLog = log.New(ioutil.Discard, "", 0)
	}
	if errLog == nil {
		errLog = log.New(ioutil.Discard, "", 0)
	}
	if dbgLog == nil {
		dbgLog = log.New(ioutil.Discard, "", 0)
	}

	im = &Importer{conf: &conf, infoLog: infoLog, errLog: errLog, dbgLog: dbgLog}

	if im.conf.Workers <= 0 {
		im.conf.Workers = runtime.NumCPU()
	}
	im.work = make(chan *Job, im.conf.Workers)
	for i := 0; i < im.conf.Workers; i = i + 1 {
		im.wgWorker.Add(1)
		go func(idx int) {
			defer im.wgWorker.Done()
			im.runWorker(idx)
		}(i)
	}
	im.jobs = newJobInventory(infoLog, errLog, dbgLog)

	infoLog.Printf("importer: started with %d worker", im.conf.Workers)
	return
}
